/*
 * @file Fast_CSC_CSR_IO.H
 *
 * @brief Definition of the routines and data to be used in Fast_CSC_CSR_IO
 *
 *  @date   16/08/2023
 *  @author Simone Bna
 */


#ifndef __Fast_CSC_CSR_IO_H__
#define __Fast_CSC_CSR_IO_H__


/**
 * Local include
 */
#include "unitTestMatVec_IO.H"
#include "Fast_CSC_CSR_IO.H"


/**
 * Include the kernels
 */
#if !defined( __DRY_RUN__ )
#include "Fast_CSC_CSR_kernels.H"
#endif


class Fast_CSC_CSR_IO:
  public unitTestMatVec
{
  private:
    
    char* m_name;
    bool m_locAllocatedInput;
    unsigned char* __restrict__ m_lineNNZ;

  public:

    /**
     * @brief Constructor for the IO object given all the parameters
     *
     * The name of the unit test (the searched file ) will be generated as follows:
     *
     *  UTname = UTdir/${name}/${prefix}_${name}_${%6.6d-index}.h5
     *
     * @param[in] UTdir  direcotry where the unit tests are stored
     * @param[in] prefix prefix to be applied to the neame of the unit tests
     * @param[in] index  index of hte unit thest
     */
    Fast_CSC_CSR_IO
    (
      const char* outputFile,
      const char* UTdir,
      const char* prefix,
      idx_k       index,
      double      tolerance
    ):
      // Inheritance
      unitTestMatVec(outputFile,UTdir,prefix,index,tolerance),
      // Members
      m_locAllocatedInput(false),
      m_lineNNZ(nullptr)
    {
      /* Construct the name of the kernel */
      m_name = (char*)malloc(32);
      memset(m_name, 0, 32);
      sprintf(m_name, "Fast_CSC_CSR");
      /* Exit point */
      return;
    };


    /**
     * @brief Destructor for the class
     */
    ~Fast_CSC_CSR_IO()
    {
      // unitTestMatVec::~unitTestMatVec();
      /* Exit point */
      return;
    };


    /**
     * @brief Constructor for the IO object.
     *        Unit test dir assumed equal to "../../data"
     *
     * The name of the unit test (the searched file ) will be generated as follows:
     *
     *  UTname = UTdir/${name}/${prefix}_${name}_${%6.6d-index}.h5
     *
     * @param[in] prefix prefix to be applied to the neame of the unit tests
     * @param[in] index  index of hte unit thest
     */
    Fast_CSC_CSR_IO
    (
      const char*  prefix,
      idx_k  index
    ):
      unitTestMatVec
      (
        "test01.db",
        "../../data/",
        prefix,
        index,
        1.0E-12
      )
    {};


    /**
     * @brief Constructor for the IO object.
     *        Index assumed equal to "0"
     *
     * The name of the unit test (the searched file ) will be generated as follows:
     *
     *  UTname = UTdir/${name}/${prefix}_${name}_${%6.6d-index}.h5
     *
     * @param[in] prefix prefix to be applied to the neame of the unit tests
     * @param[in] index  index of hte unit thest
     */
    Fast_CSC_CSR_IO
    (
      const char* prefix
    ):
      unitTestMatVec
      (
        prefix,
        0
      )
    {};


    /**
     * @brief Constructor for the IO object.
     *        Prefix assumed equal to "LidDrivenCavity3D"
     *
     * The name of the unit test (the searched file ) will be generated as follows:
     *
     *  UTname = UTdir/${name}/${prefix}_${name}_${%6.6d-index}.h5
     *
     * @param[in] prefix prefix to be applied to the neame of the unit tests
     * @param[in] index  index of hte unit thest
     */
    Fast_CSC_CSR_IO
    (
      idx_k  index
    ):
      unitTestMatVec
      (
        "LidDrivenCavity3D",
        index
      )
    {};

    /**
     * @brief Empty constructor for the IO object
     * 
     */
    Fast_CSC_CSR_IO():
      unitTestMatVec
      (
        "LidDrivenCavity3D",
        0
      )
    {};

    const char* kernelFlavour() const { return( m_name ); };

    /**
     * @brief load input data for the kernel
     */
    void 
    loadInputData( )
    {
      unitTestMatVec::loadInputData();
      /* Local data */
      if (!m_locAllocatedInput)
      {
        m_lineNNZ = static_cast<unsigned char*>(malloc(m_nCellsInp_CPU*sizeof(unsigned char)));
        
	/* Set the allocation status flag */
        m_locAllocatedInput = true;
      }
      /* Exit point */
      return;
    };


    /**
     * @brief load output data for the kernel
     */
    void 
    loadOutputData( )
    {
      unitTestMatVec::loadOutputData();
      return;
    };

    void
    computeLineNNZ()
    {
      for (idx_k i=0; i<m_nCellsInp_CPU; ++i)
      {
        m_lineNNZ[i] = 0;
      }

      for (idx_k i=0; i<m_nInternalFacesInp_CPU; ++i)
      {
        m_lineNNZ[(m_ownerInp_CPU)[i]]++;
      }
      return;
    };


    void
    Prepare()
    {
      /* Compute the number of nnz per row */
      computeLineNNZ();

#ifdef __DEBUG_VERBOSE__
      /* Save  the permuted vectors */
      fid = fopen( "blockKind.dat", "w" );
      for (idx_k i=0; i<m_permutation->nSubBlocks(); ++i)
      {
        fprintf(fid, "%d\n", m_lineNNZ[i]);
      }
      fclose(fid);
#endif
      return;
    };

    fld_k
    bandWidth() const
    {
      return( computeBandwidth
              (
                m_nCellsInp_CPU,
                m_nInternalFacesInp_CPU,
                m_ownerInp_CPU,
                m_neighInp_CPU
              )
            );
    };


    fld_k
    jumpParam() const
    {
      return( computeJumpParameter
              (
                m_nCellsInp_CPU,
                m_nInternalFacesInp_CPU,
                m_ownerInp_CPU
              )
            );
    };


    /**
     * @brief allocate data on the device
     */
    void
    allocateDevice()
    {
      unitTestMatVec::allocateDevice();
      /* Exit point */
      return;
    };


    /**
     * @brief copy data to the device
     */
    void
    copyToDevice()
    {
      unitTestMatVec::copyToDevice();
      /* Exit point */
      return;
    };


    /**
     * @brief copy data from the device
     */
    void
    copyFromDevice()
    {
      unitTestMatVec::copyFromDevice();
      /* Exit point */
      return;
    };


    /**
     * @brief Run the kernels 
     *
     * @param[in,out] profileTime time spent in the computation
     */
    void
    warmup
    (
      timeProfile& profileTime
    )
    {
      /* Start the profiling */
      m_profiler.tic();

      /* Call the kernel */
#if defined( __TEST_CPU__ ) && !defined( __DRY_RUN__ )
      matVec_Fast_CSC_CSR_CPU
      (
        m_nCellsInp_CPU,
        m_lineNNZ,
        m_neighInp_CPU,
        m_upperInp_CPU,
        m_lowerInp_CPU,
        m_psiInp_CPU,
        m_ApsiInp_CPU
      );
#endif

/* Call the CUDA kernel */
#if defined( __TEST_NATIVE_CUDA__ ) && !defined( __DRY_RUN__ )
      // Synchronize the device
      cudaDeviceSynchronize();
#endif

/* Call the OPEN_ACC kernel */
#if defined( __TEST_OPEN_ACC__ ) && !defined( __DRY_RUN__ )
      // Synchronize the device
      cudaDeviceSynchronize();
#endif

    /* Call the OPEN_MP kernel */
#if  defined( __TEST_OPEN_MP__ ) && !defined( __DRY_RUN__ )

      // Synchronize the device
      cudaDeviceSynchronize();
#endif
      /* Record the profiling time */
      m_profiler.toc( profileTime );
      /* Exit point */
      return;
    };


    /**
     * @brief Run the kernels 
     *
     * @param[in,out] profileTime time spent in the computation
     */
    void
    race
    (
      timeProfile& profileTime,
      idx_k        numberOfRepetitions
    )
    {
      /* Start the profiling */
      m_profiler.tic();

      /* Call the kernel */
#if defined( __TEST_CPU__ ) && !defined( __DRY_RUN__ )
      for ( idx_k i=0; i<numberOfRepetitions; i++ )
      {
        matVec_Fast_CSC_CSR_CPU
        (
          m_nCellsInp_CPU,
          m_lineNNZ,
          m_neighInp_CPU,
          m_upperInp_CPU,
          m_lowerInp_CPU,
          m_psiInp_CPU,
          m_ApsiInp_CPU
        );
      }
#endif

/* Call the CUDA kernel */
#if defined( __TEST_NATIVE_CUDA__ ) && !defined( __DRY_RUN__ )
      // Synchronize the device
      cudaDeviceSynchronize();
#endif

/* Call the OPEN_ACC kernel */
#if defined( __TEST_OPEN_ACC__ ) && !defined( __DRY_RUN__ )
      // Synchronize the device
      cudaDeviceSynchronize();
#endif

    /* Call the OPEN_MP kernel */
#if  defined( __TEST_OPEN_MP__ ) && !defined( __DRY_RUN__ )

      // Synchronize the device
      cudaDeviceSynchronize();
#endif
      /* Record the profiling time */
      m_profiler.toc( profileTime );
      /* Exit point */
      return;
    };


    /**
     * @brief free memory
     */
    void 
    Free()
    {
      unitTestMatVec::Free();
      /* Exit point */
      return;
    };

};

#endif


/**
 * In development phase, it is necessary to build a single unit test in order
 * to be more flexible you can enable the following define and focus on only 
 * one source file.
 */ 
#ifdef __SINGLE_MAIN_DEVEL__
#include "singleTestDevel.H"
__unitTest__( Fast_CSC_CSR_IO );
#endif
