/*
 * @file padFast8_CSC_CSR_IO.H
 *
 * @brief Definition of the routines and data to be used in padFast8_CSC_CSR_IO
 *
 *  @date   04/09/2023
 *  @author Francesco Bottau
 */


#ifndef __padFast8_CSC_CSR_IO_H__
#define __padFast8_CSC_CSR_IO_H__


/**
 * Local include
 */
#include "unitTestMatVec_IO.H"
#include "padFast8_CSC_CSR_IO.H"

/*--debug--*/
#include <fstream>
using namespace std;
/*--debug--*/

/**
 * Include the kernels
 */
#if !defined( __DRY_RUN__ )
#include "padFast_CSC_CSR_kernels.H"
#endif


class padFast8_CSC_CSR_IO:
  public unitTestMatVec
{
  private:
    
    char* m_name;
    bool m_locAllocatedInput;
    unsigned char * m_lineNNZ;
    unsigned char * m_blkNNZ;
    idx_k  m_blkRows = 8;       // Number of rows for padded block
    idx_k  m_nBlks;             // Number of complete blocks
    idx_k* m_padOwn;            // Padded owner vector
    idx_k* m_padNeigh;          // Padded neighbour vector
    idx_k  m_padNinternalfaces; // Number of internal faces after padding
    LDUoffDiag_k* m_padUpper;   // Padded Upper matrix value vector
    LDUoffDiag_k* m_padLower;   // Padded Lower matrix value vector

  public:

    /**
     * @brief Constructor for the IO object given all the parameters
     *
     * The name of the unit test (the searched file ) will be generated as follows:
     *
     *  UTname = UTdir/${name}/${prefix}_${name}_${%6.6d-index}.h5
     *
     * @param[in] UTdir  direcotry where the unit tests are stored
     * @param[in] prefix prefix to be applied to the neame of the unit tests
     * @param[in] index  index of hte unit thest
     */
    padFast8_CSC_CSR_IO
    (
      const char* outputFile,
      const char* UTdir,
      const char* prefix,
      idx_k       index,
      double      tolerance
    ):
      // Inheritance
      unitTestMatVec(outputFile,UTdir,prefix,index,tolerance),
      // Members
      m_locAllocatedInput(false),
      m_lineNNZ(nullptr)
    {
      /* Construct the name of the kernel */
      m_name = (char*)malloc(32);
      memset(m_name, 0, 32);
      sprintf(m_name, "padFast8_CSC_CSR");
      /* Exit point */
      return;
    };


    /**
     * @brief Destructor for the class
     */
    ~padFast8_CSC_CSR_IO()
    {
      // unitTestMatVec::~unitTestMatVec();
      /* Exit point */
      return;
    };


    /**
     * @brief Constructor for the IO object.
     *        Unit test dir assumed equal to "../../data"
     *
     * The name of the unit test (the searched file ) will be generated as follows:
     *
     *  UTname = UTdir/${name}/${prefix}_${name}_${%6.6d-index}.h5
     *
     * @param[in] prefix prefix to be applied to the neame of the unit tests
     * @param[in] index  index of hte unit thest
     */
    padFast8_CSC_CSR_IO
    (
      const char*  prefix,
      idx_k  index
    ):
      unitTestMatVec
      (
        "test01.db",
        "../../data/",
        prefix,
        index,
        1.0E-12
      )
    {};


    /**
     * @brief Constructor for the IO object.
     *        Index assumed equal to "0"
     *
     * The name of the unit test (the searched file ) will be generated as follows:
     *
     *  UTname = UTdir/${name}/${prefix}_${name}_${%6.6d-index}.h5
     *
     * @param[in] prefix prefix to be applied to the neame of the unit tests
     * @param[in] index  index of hte unit thest
     */
    padFast8_CSC_CSR_IO
    (
      const char* prefix
    ):
      unitTestMatVec
      (
        prefix,
        0
      )
    {};


    /**
     * @brief Constructor for the IO object.
     *        Prefix assumed equal to "LidDrivenCavity3D"
     *
     * The name of the unit test (the searched file ) will be generated as follows:
     *
     *  UTname = UTdir/${name}/${prefix}_${name}_${%6.6d-index}.h5
     *
     * @param[in] prefix prefix to be applied to the neame of the unit tests
     * @param[in] index  index of hte unit thest
     */
    padFast8_CSC_CSR_IO
    (
      idx_k  index
    ):
      unitTestMatVec
      (
        "LidDrivenCavity3D",
        index
      )
    {};

    /**
     * @brief Empty constructor for the IO object
     * 
     */
    padFast8_CSC_CSR_IO():
      unitTestMatVec
      (
        "LidDrivenCavity3D",
        0
      )
    {};

    const char* kernelFlavour() const { return( m_name ); };

    void
    computeLineNNZ(idx_k* own, idx_k nFaces)
    {
      for (idx_k i=0; i<m_nCellsInp_CPU; ++i)
      {
        m_lineNNZ[i] = 0;
      }

      for (idx_k i=0; i<nFaces; ++i)
      {
        m_lineNNZ[own[i]]++;
      }
      return;
    };

    /**
     * @brief load input data for the kernel
     */
    void 
    loadInputData( )
    {
      // local variables:
      idx_k absMaxNnz = 0;

      unitTestMatVec::loadInputData();
      m_nBlks = m_nCellsInp_CPU / m_blkRows; //n blocks with 4 rows
      
      /* Local data */
      if (!m_locAllocatedInput)
      {
        m_lineNNZ = static_cast<unsigned char*>(malloc(m_nCellsInp_CPU*sizeof(unsigned char)));
        
        computeLineNNZ(m_ownerInp_CPU, m_nInternalFacesInp_CPU);
        for(idx_k i=0; i<m_nCellsInp_CPU; ++i){
          if(m_lineNNZ[i] > absMaxNnz){
            absMaxNnz = m_lineNNZ[i];
          }
        }
        m_padOwn = reinterpret_cast<idx_k*> ( malloc( (m_nCellsInp_CPU*absMaxNnz)*sizeof(idx_k) ) );
        m_padNeigh = reinterpret_cast<idx_k*> ( malloc( (m_nCellsInp_CPU*absMaxNnz)*sizeof(idx_k) ) );
        m_padUpper = reinterpret_cast<LDUoffDiag_k*> ( malloc( (m_nCellsInp_CPU*absMaxNnz)*sizeof(LDUoffDiag_k) ) );
        m_padLower = reinterpret_cast<LDUoffDiag_k*> ( malloc( (m_nCellsInp_CPU*absMaxNnz)*sizeof(LDUoffDiag_k) ) );
        m_blkNNZ = static_cast<unsigned char*> (malloc((m_nBlks+1)*sizeof(unsigned char)));
	/* Set the allocation status flag */
        m_locAllocatedInput = true;
      }
      /* Exit point */
      return;
    };


    /**
     * @brief load output data for the kernel
     */
    void 
    loadOutputData( )
    {
      unitTestMatVec::loadOutputData();
      return;
    };

    /**
     * @brief Print matrix after padding to compare with original one
     */
    void
    plotPadMatrix( )
    {
      idx_k nnz = 0;
      idx_k l = 0, g = 0;
      LDUoffDiag_k tol = 1e-12;
      ofstream f("matrix.txt");

      for(idx_k j=0; j<m_nCellsInp_CPU; j++){
        nnz = m_lineNNZ[j];

        f << "line: " << j << ", original NNZ: " << nnz << endl;
        f << "orig: ";
        if(nnz != 0){
          for(idx_k p=j+1; p<m_neighInp_CPU[l]; p++) f << "\t";
          f << "(" << m_neighInp_CPU[l] << ", " << m_upperInp_CPU[l] << ")";
          for(idx_k s=1; s<nnz; s++ ){
            for(idx_k p=m_neighInp_CPU[l+s-1]; p<m_neighInp_CPU[l+s]; p++) f << "\t";
            f << "(" << m_neighInp_CPU[l+s] << ", " << m_upperInp_CPU[l+s] << ")";
          }
        }
        f << endl;
        l += nnz;

        f << "pad:  ";
        for(idx_k p=j+1; p<m_padNeigh[g]; p++) f << "\t";
        f << "(" << m_padNeigh[g] << ", " << m_padUpper[g] << ")";
        for(idx_k s=1; s<nnz; s++ ){
          for(idx_k p=m_padNeigh[g+s-1]; p<m_padNeigh[g+s]; p++) f << "\t";
          f << "(" << m_padNeigh[g+s] << ", " << m_padUpper[g+s] << ")";
        }
        if(nnz != 0){
          g += nnz;
        }else{
          g++;
        }
        while(m_padOwn[g] == j){
          for(idx_k p=m_padNeigh[g-1]; p<m_padNeigh[g]; p++) f << "\t";
          f << "(" << m_padNeigh[g] << ", " << m_padUpper[g] << ")";
          g++;
        }
        f << endl;
      }
      return;
    };


    void
    Prepare()
    {
      /* paddding matrix to have same number of nnz for each blocks */

      idx_k i = 0;
      idx_k l = 0;
      idx_k g = 0;
      idx_k nnz = 0;
      LDUoffDiag_k fakeZero = 0.0;
      idx_k blkRows;
      idx_k neighTmp;

      while (i < m_nCellsInp_CPU)
      {
        if(m_nCellsInp_CPU-i > m_blkRows){
          blkRows = m_blkRows;
        }else{
          blkRows = m_nCellsInp_CPU-i;
        }

        idx_k blkMaxNnz = 0;
        idx_k blkMinNnz = 0; /*--debug--*/

        // loop to find the maxNnz in the sub block
        for(idx_k j=i; j<i+blkRows; j++){
          if(m_lineNNZ[j] > blkMaxNnz) blkMaxNnz = m_lineNNZ[j];
          if(m_lineNNZ[j] < blkMinNnz) blkMinNnz = m_lineNNZ[j];  /*--debug--*/
        }

        // loop to add fake zeros to pad the subblock
        for(idx_k j=i; j<i+blkRows; j++){
          
          nnz = m_lineNNZ[j];

          if(nnz != 0){

            for(idx_k k=0; k<nnz; k++){
              m_padOwn[g + k] = j;
              m_padNeigh[g + k] = m_neighInp_CPU[l+k];
              m_padUpper[g + k] = m_upperInp_CPU[l+k];
              m_padLower[g + k] = m_lowerInp_CPU[l+k];
            }

            idx_k h = 1;
            idx_k t = 1;
            for(idx_k k=nnz; k<blkMaxNnz; k++){
              m_padOwn[g + k] = j;
              neighTmp = m_neighInp_CPU[l+nnz-1] + h;            
              if(neighTmp < m_nCellsInp_CPU){
                m_padNeigh[g + k] = neighTmp;
              }else{
                m_padNeigh[g + k] = m_neighInp_CPU[l] - t;
                t++;
              }
              m_padUpper[g + k] = fakeZero;
              m_padLower[g + k] = fakeZero;
              h++;
            }

          }else{
            idx_k h = 1;
            idx_k t = 1;
            for(idx_k k=0; k<blkMaxNnz; k++){
              m_padOwn[g + k] = j;
              neighTmp = j + h;            
              if(neighTmp < m_nCellsInp_CPU){
                m_padNeigh[g + k] = neighTmp;
              }else{
                m_padNeigh[g + k] = j - t;
                t++;
              }
              m_padUpper[g + k] = fakeZero;
              m_padLower[g + k] = fakeZero;
              h++;
            }
          }

          l += nnz;
          g += blkMaxNnz;
        }
        i += blkRows;
      }

      m_padNinternalfaces = g;

#ifdef __DEBUG_VERBOSE__
      /* write matrix in a file to compare padded mat with original one */
      plotPadMatrix( );
#endif      

      computeLineNNZ(m_padOwn, m_padNinternalfaces); //upgrade of the NNZ for line after padding

      for(idx_k n=0; n<m_nBlks+1; n++){
        m_blkNNZ[n] = m_lineNNZ[n*m_blkRows];
      }


#ifdef __DEBUG_VERBOSE__
      /* Save  the permuted vectors */
      fid = fopen( "blockKind.dat", "w" );
      for (idx_k i=0; i<m_permutation->nSubBlocks(); ++i)
      {
        fprintf(fid, "%d\n", m_lineNNZ[i]);
      }
      fclose(fid);
#endif
      return;
    };

    fld_k
    bandWidth() const
    {
      return( computeBandwidth
              (
                m_nCellsInp_CPU,
                m_padNinternalfaces,
                m_padOwn,
                m_padNeigh
              )
            );
    };


    fld_k
    jumpParam() const
    {
      return( computeJumpParameter
              (
                m_nCellsInp_CPU,
                m_padNinternalfaces,
                m_padOwn
              )
            );
    };


    /**
     * @brief allocate data on the device
     */
    void
    allocateDevice()
    {
      unitTestMatVec::allocateDevice();
      /* Exit point */
      return;
    };


    /**
     * @brief copy data to the device
     */
    void
    copyToDevice()
    {
      unitTestMatVec::copyToDevice();
      /* Exit point */
      return;
    };


    /**
     * @brief copy data from the device
     */
    void
    copyFromDevice()
    {
      unitTestMatVec::copyFromDevice();
      /* Exit point */
      return;
    };


    /**
     * @brief Run the kernels 
     *
     * @param[in,out] profileTime time spent in the computation
     */
    void
    warmup
    (
      timeProfile& profileTime
    )
    {
      /* Start the profiling */
      m_profiler.tic();

      /* Call the kernel */
#if defined( __TEST_CPU__ ) && !defined( __DRY_RUN__ )
      matVec_padFast_CSC_CSR_CPU<8>
      (
        m_nCellsInp_CPU,
        m_nBlks,
        m_blkNNZ,
        m_padNeigh,
        m_padUpper,
        m_padLower,
        m_psiInp_CPU,
        m_ApsiInp_CPU
      );
#endif

/* Call the CUDA kernel */
#if defined( __TEST_NATIVE_CUDA__ ) && !defined( __DRY_RUN__ )
      // Synchronize the device
      cudaDeviceSynchronize();
#endif

/* Call the OPEN_ACC kernel */
#if defined( __TEST_OPEN_ACC__ ) && !defined( __DRY_RUN__ )
      // Synchronize the device
      cudaDeviceSynchronize();
#endif

    /* Call the OPEN_MP kernel */
#if  defined( __TEST_OPEN_MP__ ) && !defined( __DRY_RUN__ )

      // Synchronize the device
      cudaDeviceSynchronize();
#endif
      /* Record the profiling time */
      m_profiler.toc( profileTime );
      /* Exit point */
      return;
    };


    /**
     * @brief Run the kernels 
     *
     * @param[in,out] profileTime time spent in the computation
     */
    void
    race
    (
      timeProfile& profileTime,
      idx_k        numberOfRepetitions
    )
    {
      /* Start the profiling */
      m_profiler.tic();

      /* Call the kernel */
#if defined( __TEST_CPU__ ) && !defined( __DRY_RUN__ )
      for ( idx_k i=0; i<numberOfRepetitions; i++ )
      {
        matVec_padFast_CSC_CSR_CPU<8>
        (
          m_nCellsInp_CPU,
          m_nBlks,
          m_blkNNZ,
          m_padNeigh,
          m_padUpper,
          m_padLower,
          m_psiInp_CPU,
          m_ApsiInp_CPU
        );
      }
#endif

/* Call the CUDA kernel */
#if defined( __TEST_NATIVE_CUDA__ ) && !defined( __DRY_RUN__ )
      // Synchronize the device
      cudaDeviceSynchronize();
#endif

/* Call the OPEN_ACC kernel */
#if defined( __TEST_OPEN_ACC__ ) && !defined( __DRY_RUN__ )
      // Synchronize the device
      cudaDeviceSynchronize();
#endif

    /* Call the OPEN_MP kernel */
#if  defined( __TEST_OPEN_MP__ ) && !defined( __DRY_RUN__ )

      // Synchronize the device
      cudaDeviceSynchronize();
#endif
      /* Record the profiling time */
      m_profiler.toc( profileTime );
      /* Exit point */
      return;
    };


    /**
     * @brief free memory
     */
    void 
    Free()
    {
      unitTestMatVec::Free();
      /* Exit point */
      return;
    };

};

#endif


/**
 * In development phase, it is necessary to build a single unit test in order
 * to be more flexible you can enable the following define and focus on only 
 * one source file.
 */ 
#ifdef __SINGLE_MAIN_DEVEL__
#include "singleTestDevel.H"
__unitTest__( padFast8_CSC_CSR_IO );
#endif
