/*
 * @file pad_CSC_CSR_IO.H
 *
 * @brief Definition of the routines and data to be used in pad_CSC_CSR
 *
 *  @date   11/09/2023
 *  @author Francesco Bottau
 */


#ifndef __pad_CSC_CSR_IO_H__
#define __pad_CSC_CSR_IO_H__


/**
 * Local include
 */
#include "unitTestMatVec_IO.H"

/**
 * Include the kernels
 */
#if !defined( __DRY_RUN__ )
#include "pad_CSC_CSR_kernels.H"
#endif


class pad_CSC_CSR_IO:
  public unitTestMatVec
{
  private:

    idx_k* m_rowStartInp_CPU;
    idx_k* m_rowStartOut_CPU;
    idx_k  m_blkRows = 4;       // Number of rows for padded block
    idx_k* m_padNeigh;          // Padded neighbour vector
    idx_k* m_padOwn;            // Padded owner vector
    LDUoffDiag_k* m_padUpper;   // Padded Upper matrix value vector
    LDUoffDiag_k* m_padLower;   // Padded Lower matrix value vector
    idx_k* m_lineNNZ;           // vector of number of non zeros per line
    idx_k  m_padNinternalfaces; // number of internal faces after padding

void 
ComputeRowStart
(
    const idx_k                        nCells,
    const idx_k                        nInternalFaces,
    const idx_k* const __MY_RESTRICT__ owner,
          idx_k* const __MY_RESTRICT__ rowStart
)
{
  // Local variables
  idx_k iFace;
  idx_k nOwnStart;
  idx_k curOwn;
  idx_k idx;
  // Compute row start
  nOwnStart   = 0;
  rowStart[0] = 0;
  idx         = 1;
  for ( iFace=0; iFace<nInternalFaces; iFace++ )
  {
    curOwn = owner[iFace];
    if ( curOwn > nOwnStart )
    {
      while ( idx <= curOwn )
      {
        rowStart[idx] = iFace;
        idx++;
      };
      nOwnStart = curOwn;
    };
  };
  // Fill the remaining part of the ros start vector
  for (iFace=idx;iFace<nCells+1;iFace++)
  {
    rowStart[iFace]=nInternalFaces;
  };
  // Exit point
  return;
};

  public:

    /**
     * @brief Constructor for the IO object given all the parameters
     *
     * The name of the unit test (the searched file ) will be generated as follows:
     *
     *  UTname = UTdir/${name}/${prefix}_${name}_${%6.6d-index}.h5
     *
     * @param[in] UTdir  direcotry where the unit tests are stored
     * @param[in] prefix prefix to be applied to the neame of the unit tests
     * @param[in] index  index of hte unit thest
     */
    pad_CSC_CSR_IO
    (
      const char* outputFile,
      const char* UTdir,
      const char* prefix,
      idx_k       index,
      double      tolerance
    ):
      unitTestMatVec(outputFile,UTdir,prefix,index,tolerance)
     ,m_rowStartInp_CPU(nullptr)
     ,m_rowStartOut_CPU(nullptr)
    {};


    /**
     * @brief Destructor for the class
     */
    ~pad_CSC_CSR_IO()
    {
      Free();
      /* Exit point */
      return;
    };


    /**
     * @brief Constructor for the IO object.
     *        Unit test dir assumed equal to "../../data"
     *
     * The name of the unit test (the searched file ) will be generated as follows:
     *
     *  UTname = UTdir/${name}/${prefix}_${name}_${%6.6d-index}.h5
     *
     * @param[in] prefix prefix to be applied to the neame of the unit tests
     * @param[in] index  index of hte unit thest
     */
    pad_CSC_CSR_IO
    (
      const char*  prefix,
      idx_k  index
    ):
      unitTestMatVec
      (
        "test01.db",
        "../../data/",
        prefix,
        index,
        1.0E-12
      )
    {};


    /**
     * @brief Constructor for the IO object.
     *        Index assumed equal to "0"
     *
     * The name of the unit test (the searched file ) will be generated as follows:
     *
     *  UTname = UTdir/${name}/${prefix}_${name}_${%6.6d-index}.h5
     *
     * @param[in] prefix prefix to be applied to the neame of the unit tests
     * @param[in] index  index of hte unit thest
     */
    pad_CSC_CSR_IO
    (
      const char* prefix
    ):
      unitTestMatVec
      (
        prefix,
        0
      )
    {};


    /**
     * @brief Constructor for the IO object.
     *        Prefix assumed equal to "LidDrivenCavity3D"
     *
     * The name of the unit test (the searched file ) will be generated as follows:
     *
     *  UTname = UTdir/${name}/${prefix}_${name}_${%6.6d-index}.h5
     *
     * @param[in] prefix prefix to be applied to the neame of the unit tests
     * @param[in] index  index of hte unit thest
     */
    pad_CSC_CSR_IO
    (
      idx_k  index
    ):
      unitTestMatVec
      (
        "LidDrivenCavity3D",
        index
      )
    {};

    /**
     * @brief Empty constructor for the IO object
     * 
     */
    pad_CSC_CSR_IO():
      unitTestMatVec
      (
        "LidDrivenCavity3D",
        0
      )
    {};


    const char* kernelFlavour() const { return( "pad_CSC_CSR" ); };

    /**
     * @brief compute NNZ per each line
     */
    void
    computeLineNNZ()
    {
      for (idx_k i=0; i<m_nCellsInp_CPU; ++i)
      {
        m_lineNNZ[i] = 0;
      }

      for (idx_k i=0; i<m_nInternalFacesInp_CPU; ++i)
      {
        m_lineNNZ[(m_ownerInp_CPU)[i]]++;
      }

      m_lineNNZ[m_nCellsInp_CPU] = 22;

      return;
    };

    /**
     * @brief load input data for the kernel
     */
    void 
    loadInputData( )
    {
      // local variables:
      idx_k absMaxNnz = 0;

      // load global variables:
      unitTestMatVec::loadInputData();

      // auxiliary functions
      m_lineNNZ = static_cast<idx_k*>(malloc((m_nCellsInp_CPU+1)*sizeof(idx_k)));
      computeLineNNZ();
      for(idx_k i=0; i<m_nCellsInp_CPU; ++i){
        if(m_lineNNZ[i] > absMaxNnz){
          absMaxNnz = m_lineNNZ[i];
        }
      }
      
      // Allocate memory
      m_padOwn = reinterpret_cast<idx_k*> ( malloc( (m_nCellsInp_CPU*absMaxNnz)*sizeof(idx_k) ) );
      m_padNeigh = reinterpret_cast<idx_k*> ( malloc( (m_nCellsInp_CPU*absMaxNnz)*sizeof(idx_k) ) );
      m_padUpper = reinterpret_cast<LDUoffDiag_k*> ( malloc( (m_nCellsInp_CPU*absMaxNnz)*sizeof(LDUoffDiag_k) ) );
      m_padLower = reinterpret_cast<LDUoffDiag_k*> ( malloc( (m_nCellsInp_CPU*absMaxNnz)*sizeof(LDUoffDiag_k) ) );
      m_rowStartInp_CPU = reinterpret_cast<idx_k*> ( malloc( (m_nCellsInp_CPU+1)*sizeof(idx_k) ) );
      /* Exit point */
      return;
    };


    /**
     * @brief load output data for the kernel
     */
    void 
    loadOutputData( )
    {
      unitTestMatVec::loadOutputData();
      m_rowStartOut_CPU = reinterpret_cast<idx_k*> ( malloc( (m_nCellsInp_CPU+1)*sizeof(idx_k) ) );
      /* Exit point */
      return;
    };

    /**
     * @brief Prepare data for the specific kernel
     */
    void
    Prepare()
    {
      idx_k i = 0;
      idx_k l = 0;
      idx_k g = 0;
      idx_k nnz = 0;
      LDUoffDiag_k fakeZero = 0.0;
      idx_k blkRows;
      idx_k neighTmp;

      while (i < m_nCellsInp_CPU)
      {
        if(m_nCellsInp_CPU-i > m_blkRows){
          blkRows = m_blkRows;
        }else{
          blkRows = m_nCellsInp_CPU-i;
        }

        idx_k blkMaxNnz = 0;

        // loop to find the maxNnz in the sub block
        for(idx_k j=i; j<i+blkRows; j++){
          if(m_lineNNZ[j] > blkMaxNnz) blkMaxNnz = m_lineNNZ[j];
        }

        // loop to add fake zeros to pad the subblock
        for(idx_k j=i; j<i+blkRows; j++){
          
          nnz = m_lineNNZ[j];

          for(idx_k k=0; k<nnz; k++){
            m_padOwn[g + k] = j;
            m_padNeigh[g + k] = m_neighInp_CPU[l+k];
            m_padUpper[g + k] = m_upperInp_CPU[l+k];
            m_padLower[g + k] = m_lowerInp_CPU[l+k];
          }

          idx_k h = 1;
          idx_k t = 1;
          for(idx_k k=nnz; k<blkMaxNnz; k++){
            m_padOwn[g + k] = j;
            neighTmp = m_neighInp_CPU[l+nnz-1] + h;            
            if(neighTmp < m_nCellsInp_CPU){
              m_padNeigh[g + k] = m_neighInp_CPU[l+nnz-1] + h;
            }else if(m_neighInp_CPU[l] == 0){
              m_padNeigh[g + k] = m_neighInp_CPU[i] + h;
            }else{
              m_padNeigh[g + k] = m_neighInp_CPU[l] - t;
              t++;
            }
            m_padUpper[g + k] = fakeZero;
            m_padLower[g + k] = fakeZero;
            h++;
          }
          
          l += nnz;
          g += blkMaxNnz;

        }

        i += blkRows;
        
      }

      m_padNinternalfaces = g;

      /* Compute the rowstart vector needd for CSR */
      ComputeRowStart( m_nCellsInp_CPU, m_padNinternalfaces, m_padOwn, m_rowStartInp_CPU );
      ComputeRowStart( m_nCellsInp_CPU, m_padNinternalfaces, m_padOwn, m_rowStartOut_CPU );
      /* Exit point */

      return;
    };


    /**
     * @brief allocate data on the device
     */
    void
    allocateDevice()
    {
#if defined(__TEST_NATIVE_CUDA__) || defined(__TEST_OPEN_ACC__) || defined(__TEST_OPEN_MP__)
      if ( !m_allocatedDevice )
      {
        // Allocate memory
        cudaMalloc( (void**)&m_rowStart_Device, (m_nCellsInp_CPU+1)*sizeof(idx_k) );
        // Synchronize the device
        cudaDeviceSynchronize();
      }
      else
      {
        printf( "ERROR :: data already allocated:: %s, %s, %d\n", __FUNCTION__, __FILE__, __LINE__ );
        abort();
      };
#endif
      unitTestMatVec::allocateDevice();
      /* Exit point */
      return;
    };


    /**
     * @brief copy data to the device
     */
    void
    copyToDevice()
    {
      unitTestMatVec::copyToDevice();
#if defined(__TEST_NATIVE_CUDA__) || defined(__TEST_OPEN_ACC__) || defined(__TEST_OPEN_MP__)
      if ( m_allocatedInput && m_allocatedDevice )
      {
        // Allocate memory
        cudaMemcpy( m_rowStart_Device, m_rowStartInp_CPU, (m_nCellsInp_CPU+1)*sizeof(idx_k), cudaMemcpyHostToDevice );
        // Synchronize the device
        cudaDeviceSynchronize();
      }
      else
      {
        printf( "ERROR :: data are not allocated:: %s, %s, %d\n", __FUNCTION__, __FILE__, __LINE__ );
        abort();
      };
#endif
      /* Exit point */
      return;
    };


    /**
     * @brief copy data from the device
     */
    void
    copyFromDevice()
    {
      unitTestMatVec::copyFromDevice();
#if defined(__TEST_NATIVE_CUDA__) || defined(__TEST_OPEN_ACC__) || defined(__TEST_OPEN_MP__)
      if ( m_allocatedInput && m_allocatedDevice )
      {
        // Allocate memory
        cudaMemcpy( m_rowStartInp_CPU, m_rowStart_Device, (m_nCellsInp_CPU+1)*sizeof(idx_k), cudaMemcpyDeviceToHost );
        // Synchronize the device
        cudaDeviceSynchronize();
      }
      else
      {
        printf( "ERROR :: data are not allocated:: %s, %s, %d\n", __FUNCTION__, __FILE__, __LINE__ );
        abort();
      };
#endif
      /* Exit point */
      return;
    };


    /**
     * @brief Run the kernels 
     *
     * @param[in,out] profileTime time spent in the computation
     */
    void
    warmup
    (
      timeProfile& profileTime
    )
    {
      /* Start the profiling */
      m_profiler.tic();
      /* Call the kernel */
#if defined( __TEST_CPU__ ) && !defined( __DRY_RUN__ )
    matVec_pad_CSC_CSR_CPU
    (
      m_nCellsInp_CPU,
      m_padNeigh,
      m_rowStartInp_CPU,
      m_padUpper,
      m_padLower,
      m_psiInp_CPU,
      m_ApsiInp_CPU
    );
#endif

/* Call the CUDA kernel */
#if defined( __TEST_NATIVE_CUDA__ ) && !defined( __DRY_RUN__ )
      int nThreadsPerBlock = 256;
      int nBlocks = (m_nInternalFacesInp_CPU+255)/256;
      matVec_pad_CSC_CSR_nativeCUDA<<<nBlocks,nThreadsPerBlock>>>
      (
        m_nCellsInp_CPU,
        m_nFacesInp_CPU,
        m_nInternalFacesInp_CPU,
        m_cellsStrideInp_CPU,
        m_facesStrideInp_CPU,
        m_iCompInp_CPU,
        m_owner_Device,
        m_neigh_Device,
        m_rowStart_Device,
        m_upper_Device,
        m_lower_Device,
        m_psi_Device,
        m_Apsi_Device
      );
      // Synchronize the device
      cudaDeviceSynchronize();
#endif

/* Call the OPEN_ACC kernel */
#if defined( __TEST_OPEN_ACC__ ) && !defined( __DRY_RUN__ )
      matVec_pad_CSC_CSR_openACC
      (
        m_nCellsInp_CPU,
        m_nFacesInp_CPU,
        m_nInternalFacesInp_CPU,
        m_cellsStrideInp_CPU,
        m_facesStrideInp_CPU,
        m_iCompInp_CPU,
        m_owner_Device,
        m_neigh_Device,
        m_rowStart_Device,
        m_upper_Device,
        m_lower_Device,
        m_psi_Device,
        m_Apsi_Device
      );
      // Synchronize the device
      cudaDeviceSynchronize();
#endif

    /* Call the OPEN_MP kernel */
#if  defined( __TEST_OPEN_MP__ ) && !defined( __DRY_RUN__ )
      matVec_pad_CSC_CSR_openMP
      (
        m_nCellsInp_CPU,
        m_nFacesInp_CPU,
        m_nInternalFacesInp_CPU,
        m_cellsStrideInp_CPU,
        m_facesStrideInp_CPU,
        m_iCompInp_CPU,
        m_owner_Device,
        m_neigh_Device,
        m_rowStart_Device,
        m_upper_Device,
        m_lower_Device,
        m_psi_Device,
        m_Apsi_Device 
      );
      // Synchronize the device
      cudaDeviceSynchronize();
#endif
      /* Record the profiling time */
      m_profiler.toc( profileTime );
      /* Exit point */
      return;
    };

    /**
     * @brief Run the kernels 
     *
     * @param[in,out] profileTime time spent in the computation
     */
    void
    race
    (
      timeProfile& profileTime,
      idx_k        numberOfRepetitions
    )
    {
      /* Start the profiling */
      m_profiler.tic();
      /* Call the kernel */
#if defined( __TEST_CPU__ ) && !defined( __DRY_RUN__ )
      for ( idx_k i=0; i<numberOfRepetitions; ++i )
      {
        matVec_pad_CSC_CSR_CPU
        (
          m_nCellsInp_CPU,
          m_padNeigh,
          m_rowStartInp_CPU,
          m_padUpper,
          m_padLower,
          m_psiInp_CPU,
          m_ApsiInp_CPU
        );
      }
#endif

/* Call the CUDA kernel */
#if defined( __TEST_NATIVE_CUDA__ ) && !defined( __DRY_RUN__ )
      int nThreadsPerBlock = 256;
      int nBlocks = (m_nInternalFacesInp_CPU+255)/256;
      matVec_pad_CSC_CSR_nativeCUDA<<<nBlocks,nThreadsPerBlock>>>
      (
        m_nCellsInp_CPU,
        m_nFacesInp_CPU,
        m_nInternalFacesInp_CPU,
        m_cellsStrideInp_CPU,
        m_facesStrideInp_CPU,
        m_iCompInp_CPU,
        m_owner_Device,
        m_neigh_Device,
        m_rowStart_Device,
        m_upper_Device,
        m_lower_Device,
        m_psi_Device,
        m_Apsi_Device
      );
      // Synchronize the device
      cudaDeviceSynchronize();
#endif

/* Call the OPEN_ACC kernel */
#if defined( __TEST_OPEN_ACC__ ) && !defined( __DRY_RUN__ )
      matVec_pad_CSC_CSR_openACC
      (
        m_nCellsInp_CPU,
        m_nFacesInp_CPU,
        m_nInternalFacesInp_CPU,
        m_cellsStrideInp_CPU,
        m_facesStrideInp_CPU,
        m_iCompInp_CPU,
        m_owner_Device,
        m_neigh_Device,
        m_rowStart_Device,
        m_upper_Device,
        m_lower_Device,
        m_psi_Device,
        m_Apsi_Device
      );
      // Synchronize the device
      cudaDeviceSynchronize();
#endif

    /* Call the OPEN_MP kernel */
#if  defined( __TEST_OPEN_MP__ ) && !defined( __DRY_RUN__ )
      matVec_pad_CSC_CSR_openMP
      (
        m_nCellsInp_CPU,
        m_nFacesInp_CPU,
        m_nInternalFacesInp_CPU,
        m_cellsStrideInp_CPU,
        m_facesStrideInp_CPU,
        m_iCompInp_CPU,
        m_owner_Device,
        m_neigh_Device,
        m_rowStart_Device,
        m_upper_Device,
        m_lower_Device,
        m_psi_Device,
        m_Apsi_Device
      );
      // Synchronize the device
      cudaDeviceSynchronize();
#endif
      /* Record the profiling time */
      m_profiler.toc( profileTime );
      /* Exit point */
      return;
    };



    /**
     * @brief free memory
     */
    void 
    Free()
    {
      if ( m_allocatedInput )
      {
        free(m_rowStartInp_CPU);
      };
      if ( m_allocatedOutput )
      {
        free(m_rowStartOut_CPU);
      };
#if defined(__TEST_NATIVE_CUDA__) || defined(__TEST_OPEN_ACC__) || defined(__TEST_OPEN_MP__)
      if ( m_allocatedDevice )
      {
        cudaFree(m_rowStart_Device);
      };
#endif
      unitTestMatVec::Free();
      /* Exit point */
      return;
    };

};

#endif





/**
 * In development phase, it is necessary to build a single unit test in order
 * to be more flexible you can enable the following define and focus on only 
 * one source file.
 */ 
#ifdef __SINGLE_MAIN_DEVEL__
#include "singleTestDevel.H"
__unitTest__( pad_CSC_CSR_IO );
#endif
