#!/usr/bin/python3

import sqlite3
import os
import numpy as np
import matplotlib.pyplot as plt
import argparse

# DataBase TABLE CONTENT
# NAME: profiling
# 1.  date text, 
# 2.  machine text, 
# 3.  compiler text, 
# 4.  flags text, 
# 5.  hostname text, 
# 6.  prefix text, 
# 7.  isRCM int, 
# 8.  nRows int,
# 9.  procID int,
# 10. kernel text,
# 11. kernelFlavour text,
# 12. caseIndex int,
# 13. mpiIndex int,
# 14. mpiSize int,
# 16. nRepetitions int,
# 17. bandWidth double,
# 18. jumpParam double,
# 19. warmupElapsedTime double,
# 20. warmupClockTime double,
# 21. warmupCudaTime double,
# 22. raceElapsedTime double,
# 23. raceClockTime double,
# 24. raceCudaTime double,
## instrStlOps
# 25. stallMemWrite double,
# 26. stallRes double,
# 27. totalCycles double,
# 28. dpOperation double,
# 29. dpVectorOperations double
## instCache
# 25. L1_DCM double,
# 26. L1_ICM double,
# 27. L2_DCM double,
# 28. L2_ICM double,
# 29. L3_TCM double
## instrIns
# 25. loadInstructions double,
# 26. storeInstruction double,
# 27. branchInstraction double,
# 28. branchCondIns double,
# 29. branchMissPred double,
# 30. totalInstructions double


def sortingFunc(e):
    value = 0
    if "BlockRowPerm" in e:
        value = int(e.split("_")[2]) #.split(".")[0])
    return value

def plotGraph(quantity, quantityType, path, std = None):
    fig, ax = plt.subplots()
    if std:
        ax.bar(list(quantity.keys()), list(quantity.values()), width=0.5, yerr=std)
    else:
        ax.bar(list(quantity.keys()), list(quantity.values()), width=0.5)
    plt.xticks(range(len(quantity)), list(quantity.keys()), rotation='vertical')
    ax.set_ylabel(f"{quantityType} [s]")
    ax.set_title(f"{mat} - {nRow}: {quantityType} comparison ")
    fig.savefig(os.path.join(path, f"{mat}_{nRow}-{quantityType}_comparison.png"), bbox_inches='tight')

if __name__ == '__main__':
    parser = argparse.ArgumentParser(prog="plotPerformance", allow_abbrev=False)
    parser.add_argument("dataPath", type=str, help="Data path")
    parser.add_argument("--compiler", type=str, nargs=1, help="compiler used for the test")
    parser.add_argument("--instrumented", type=int, default=0, required=False, help="1: stlOps, 2: cache, 3: ins")
    parser.add_argument("--nRows", type=str, nargs='+', default=["10krows"], help="Number of rows for which the analysis has to be performed")
    parser.add_argument("--postPath", type=str, nargs=1, help="Path where to save images")

    args = parser.parse_args()

    dataPath = vars(args)["dataPath"]
    compiler = vars(args)["compiler"][0]
    instr = vars(args)["instrumented"]
    nRows = vars(args)["nRows"][:]
    postPath = vars(args)["postPath"][0]

    if instr > 3:
        print("Invalid instrumentation required.")
        exit()

    matsType = os.listdir(dataPath)
    matsType = list(filter(lambda x : os.path.isdir(os.path.join(dataPath, x)), matsType)) # ["PLMmats-RCM", "HEXmats-RCM", "SHMmats-RCM", "PLMmatsCFmesh-RCM" , "TETmatsCFmesh-RCM"]
    matsType.sort()
    print(matsType)

    raceTimeMaxMean = dict()    # Dictionary of the mean of between the maximum race time used by the MPIprocess
    raceTimeMaxMin_err = dict() # Dictionary of the difference between raceTimeMaxMean and miinimum value (computed for graph error)
    raceTimeMaxMax = dict()     # Dictionary of the maximum of between the maximum race time used by the MPIprocess
    raceTimeMaxMax_err = dict() # Dictionary of the difference between raceTimeMaxMean and maximum value (computed for graph error)
    warmUpTimeMaxMean = dict()  # Dictionary of the mean of between the maximum warm up time used by the MPIprocess
    warmUpTimeMaxMin_err = dict() # Dictionary of the difference between warmUpTimeMaxMean and miinimum value (computed for graph error)
    warmUpTimeMaxMax = dict()   # Dictionary of the maximum of between the maximum warm up time used by the MPIprocess
    warmUpTimeMaxMax_err = dict() # Dictionary of the difference between warmUpTimeMaxMean and maximum value (computed for graph error)

    if instr == 1: #stlOps
        stlMemWrt =  dict()
        stlRes = dict()
        totCyc =  dict() 
        dpOps = dict()
        dpVecOps = dict()
    elif instr == 2: #cache
        l1dcm = dict()
        l1icm = dict()
        l2dcm = dict()
        l2icm = dict()
        l3tcm = dict()
    elif instr == 3: #ins
        ldIns = dict()
        stIns = dict()
        brIns = dict()
        brConIns = dict()
        brMiss = dict()
        totIns = dict()

    barWidth = 1/(len(matsType)+1)
    aspectRatio = 0.75

    for nRow in nRows:
        print(f"processing number of rows: {nRow}")
        multiplier = 0

        ## Initialization of figures
        # 1. Figure initialization raceTime Mean Max graph
        fig1, ax1 = plt.subplots(figsize=[12, 4])
        ax1.set_ylabel("Time [ms]")
        # 2. Figure initialization raceTime Max Max graph
        fig2, ax2 = plt.subplots(figsize=[12, 4])
        ax2.set_ylabel("Time [ms]")
        # 3. Figure initialization warmUpTime Mean Max graph
        fig3, ax3 = plt.subplots(figsize=[12, 4])
        ax3.set_ylabel("Time [ms]")
        # 4. Figure initialization warmUpTime Max Max graph
        fig4, ax4 = plt.subplots(figsize=[12, 4])
        ax4.set_ylabel("Time [ms]")
        if instr == 1:
            # 5. Figure initialization stalMemWrite graph
            fig5, ax5 = plt.subplots(figsize=[12, 4])
            ax5.set_ylabel("Memory write stall")
            # 6. Figure initialization stallRes graph
            fig6, ax6 = plt.subplots(figsize=[12, 4])
            ax6.set_ylabel("stal on resources")
            # 7. Figure initialization totalCycles graph
            fig7, ax7 = plt.subplots(figsize=[12, 4])
            ax7.set_ylabel("total cycles")
            # 8. Figure initialization dpOps graph
            fig8, ax8 = plt.subplots(figsize=[12, 4])
            ax8.set_ylabel("double precision operations")
            # 9. Figure initialization dpVectorOperations graph
            fig9, ax9 = plt.subplots(figsize=[12, 4])
            ax9.set_ylabel("double precision vectorial operations")
        elif instr == 2:
            # 5. Figure initialization L1_DCM graph
            fig5, ax5 = plt.subplots(figsize=[12, 4])
            ax5.set_ylabel("L1 data cache Misses")
            # 6. Figure initialization L1_ICM graph
            fig6, ax6 = plt.subplots(figsize=[12, 4])
            ax6.set_ylabel("L1 instruction Cache Misses")
            # 7. Figure initialization L2_DCM graph
            fig7, ax7 = plt.subplots(figsize=[12, 4])
            ax7.set_ylabel("L2 data Cache Misses")
            # 8. Figure initialization L2_ICM graph
            fig8, ax8 = plt.subplots(figsize=[12, 4])
            ax8.set_ylabel("L2 instructions Cache Misses")
            # 9. Figure initialization L3_TCM graph
            fig9, ax9 = plt.subplots(figsize=[12, 4])
            ax9.set_ylabel("L3 total Cache Misses")
        elif instr == 3:
            # 5. Figure initialization loadInstructions graph
            fig5, ax5 = plt.subplots(figsize=[12, 4])
            ax5.set_ylabel("Loaded instructions")
            # 6. Figure initialization storeInstruction graph
            fig6, ax6 = plt.subplots(figsize=[12, 4])
            ax6.set_ylabel("Stored instructions")
            # 7. Figure initialization branchInstraction graph
            fig7, ax7 = plt.subplots(figsize=[12, 4])
            ax7.set_ylabel("Branch instructions")
            # 8. Figure initialization branchCondIns graph
            fig8, ax8 = plt.subplots(figsize=[12, 4])
            ax8.set_ylabel("branch conditional instructions")
            # 9. Figure initialization branchMissPred graph
            fig9, ax9 = plt.subplots(figsize=[12, 4])
            ax9.set_ylabel("Branch misspredictions")
            # 10. Figure initialization totalInstructions graph
            fig10, ax10 = plt.subplots(figsize=[12, 4])
            ax10.set_ylabel("Total instructions")

        for mat in matsType:
            print(f"processing matrix type: {mat}")  
            nRowPath = os.path.join(dataPath, mat, nRow)
            if instr == 1:
                dataBases = list(filter(lambda file: "instrStlOps.sqlite.db" in file, os.listdir(nRowPath)))
            elif instr == 2:
                dataBases = list(filter(lambda file: "instrCache.sqlite.db" in file, os.listdir(nRowPath)))
            elif instr == 3:
                dataBases = list(filter(lambda file: "instrIns.sqlite.db" in file, os.listdir(nRowPath)))
            else:
                dataBases = list(filter(lambda file: ".sqlite.db" in file, os.listdir(nRowPath)))
            dataBases.sort()
            dataBases.sort(key=sortingFunc)
            for db in dataBases:
                table = sqlite3.connect(os.path.join(nRowPath, db))
                cur = table.cursor()
                if instr == 1:
                    alg = db.replace("timings_", "")
                    alg = alg.strip("_instrStlOps.sqlite.db")
                elif instr == 2:
                    alg = db.replace("timings_", "").strip("_instrCache.sqlite.db")
                elif instr == 3:
                    alg = db.replace("timings_", "").strip("_instrIns.sqlite.db")
                else:
                    alg = db.replace("timings_", "").strip(".sqlite.db")

                print(f"Processing DataBase: {db}, Algorithm = {alg}")
                dataSet = list(cur.execute("SELECT compiler, procID, raceClockTime, warmupClockTime FROM profiling"))
                rTimeTmp = []
                wuTimeTmp = []
                raceTime = []
                wuTime = []
                procID = dataSet[0][1]
                print(f"processing processor{procID}")
                for i in range(0, len(dataSet)):
                    if dataSet[i][0] == compiler:
                        if dataSet[i][1] == procID:
                            rTimeTmp.append(dataSet[i][2]*1000)     # Conversion in [ms]
                            wuTimeTmp.append(dataSet[i][3]*1000)    # Conversion in [ms]
                        else:
                            raceTime.append(max(rTimeTmp))          # considero solo il massimo dei 48 processi MPI
                            wuTime.append(max(wuTimeTmp))           # considero solo il massimo dei 48 processi MPI
                            timeTmp = []
                            wuTimeTmp = []
                            procID = dataSet[i+1][1]
                            print(f"processing processor{procID}")
                raceTimeMaxMean[alg] = np.mean(raceTime)    # prendo la media tra le scomposizioni
                raceTimeMaxMin_err[alg] = abs(raceTimeMaxMean[alg] - min(raceTime))
                raceTimeMaxMax[alg] = max(raceTime, default=0)
                raceTimeMaxMax_err[alg] = abs(raceTimeMaxMax[alg] - raceTimeMaxMean[alg])
                warmUpTimeMean = np.mean(wuTimeTmp)
                wuTime = list(filter(lambda x : abs(x-warmUpTimeMean) < 0.75*warmUpTimeMean, wuTime)) # filtering saples too far from the mean to simplify the graph reading
                warmUpTimeMaxMean[alg] = np.mean(wuTime)
                warmUpTimeMaxMin_err[alg] = abs(warmUpTimeMaxMean[alg] - min(wuTime, default=0))
                warmUpTimeMaxMax[alg] = max(wuTime, default=0)
                warmUpTimeMaxMax_err[alg] = abs(warmUpTimeMaxMax[alg] - warmUpTimeMaxMean[alg])

                if instr == 1:
                    dataSet = list(cur.execute("SELECT compiler, stallMemWrite, stallRes, totalCycles, dpOperation, dpVectorOperations FROM profiling"))
                    stlMemWrt_tmp = []
                    stlRes_tmp = []
                    totCyc_tmp = []
                    dpOps_tmp = []
                    dpVecOps_tmp = []
                    for i in range(0, len(dataSet)):
                        if dataSet[i][0] == compiler:
                            stlMemWrt_tmp.append(dataSet[i][1])
                            stlRes_tmp.append(dataSet[i][2])
                            totCyc_tmp.append(dataSet[i][3])
                            dpOps_tmp.append(dataSet[i][4])
                            dpVecOps_tmp.append(dataSet[i][5])
                    stlMemWrt[alg] = np.mean(stlMemWrt_tmp)
                    stlRes[alg] = np.mean(stlRes_tmp)
                    totCyc[alg] = np.mean(totCyc_tmp)
                    dpOps[alg] = np.mean(dpOps_tmp)
                    dpVecOps[alg] = np.mean(dpVecOps_tmp)

                elif instr == 2:
                    dataSet = list(cur.execute("SELECT compiler, L1_DCM, L1_ICM, L2_DCM, L2_ICM, L3_TCM FROM profiling"))
                    l1dcm_tmp = []
                    l1icm_tmp = []
                    l2dcm_tmp = []
                    l2icm_tmp = []
                    l3tcm_tmp = []
                    for i in range(0, len(dataSet)):
                        if dataSet[i][0] == compiler:
                            l1dcm_tmp.append(dataSet[i][1])
                            l1icm_tmp.append(dataSet[i][2])
                            l2dcm_tmp.append(dataSet[i][3])
                            l2icm_tmp.append(dataSet[i][4])
                            l3tcm_tmp.append(dataSet[i][5])
                    l1dcm[alg] = np.mean(l1dcm_tmp)
                    l1icm[alg] = np.mean(l1icm_tmp)
                    l2dcm[alg] = np.mean(l2dcm_tmp)
                    l2icm[alg] = np.mean(l2icm_tmp)
                    l3tcm[alg] = np.mean(l3tcm_tmp)

                elif instr == 3:
                    dataSet = list(cur.execute("SELECT compiler, loadInstructions, storeInstruction, branchInstraction, branchCondIns, branchMissPred, totalInstructions FROM profiling"))
                    ldIns_tmp = []
                    stIns_tmp = []
                    brIns_tmp = []
                    brConIns_tmp = []
                    brMiss_tmp = []
                    totIns_tmp = []
                    for i in range(0, len(dataSet)):
                        if dataSet[i][0] == compiler:
                            ldIns_tmp.append(dataSet[i][1])
                            stIns_tmp.append(dataSet[i][2])
                            brIns_tmp.append(dataSet[i][3])
                            brConIns_tmp.append(dataSet[i][4])
                            brMiss_tmp.append(dataSet[i][5])
                            totIns_tmp.append(dataSet[i][6])
                    ldIns[alg] = np.mean(ldIns_tmp)
                    stIns[alg] = np.mean(stIns_tmp)
                    brIns[alg] = np.mean(brIns_tmp)
                    brConIns[alg] = np.mean(brConIns_tmp)
                    brMiss[alg] = np.mean(brMiss_tmp)
                    totIns[alg] = np.mean(totIns_tmp)

                table.close()

            # plotting collective graphs
            x = np.arange(len(raceTimeMaxMean.keys()))
            raceTimeErr = [list(raceTimeMaxMin_err.values()), list(raceTimeMaxMax_err.values())]
            warmUpTimeErr = [list(warmUpTimeMaxMin_err.values()), list(warmUpTimeMaxMax_err.values())]
            ax1.bar(x + multiplier*barWidth, list(raceTimeMaxMean.values()), width=barWidth, yerr=raceTimeErr, label=mat)
            ax2.bar(x + multiplier*barWidth, list(raceTimeMaxMax.values()), width=barWidth, label=mat)
            ax3.bar(x + multiplier*barWidth, list(warmUpTimeMaxMean.values()), width=barWidth, yerr=warmUpTimeErr, label=mat)
            ax4.bar(x + multiplier*barWidth, list(warmUpTimeMaxMax.values()), width=barWidth, label=mat)
            if instr == 1:
                x = np.arange(len(stlMemWrt.keys()))
                ax5.bar(x + multiplier*barWidth, list(stlMemWrt.values()), width=barWidth, label=mat)
                ax6.bar(x + multiplier*barWidth, list(stlRes.values()), width=barWidth, label=mat)
                ax7.bar(x + multiplier*barWidth, list(totCyc.values()), width=barWidth, label=mat)
                ax8.bar(x + multiplier*barWidth, list(dpOps.values()), width=barWidth, label=mat)
                ax9.bar(x + multiplier*barWidth, list(dpVecOps.values()), width=barWidth, label=mat)
            elif instr == 2:
                x = np.arange(len(l1dcm.keys()))
                ax5.bar(x + multiplier*barWidth, list(l1dcm.values()), width=barWidth, label=mat)
                ax6.bar(x + multiplier*barWidth, list(l1icm.values()), width=barWidth, label=mat)
                ax7.bar(x + multiplier*barWidth, list(l2dcm.values()), width=barWidth, label=mat)
                ax8.bar(x + multiplier*barWidth, list(l2icm.values()), width=barWidth, label=mat)
                ax9.bar(x + multiplier*barWidth, list(l3tcm.values()), width=barWidth, label=mat)
            elif instr == 3:
                x = np.arange(len(ldIns.keys()))
                ax5.bar(x + multiplier*barWidth, list(ldIns.values()), width=barWidth, label=mat)
                ax6.bar(x + multiplier*barWidth, list(stIns.values()), width=barWidth, label=mat)
                ax7.bar(x + multiplier*barWidth, list(brIns.values()), width=barWidth, label=mat)
                ax8.bar(x + multiplier*barWidth, list(brConIns.values()), width=barWidth, label=mat)
                ax9.bar(x + multiplier*barWidth, list(brMiss.values()), width=barWidth, label=mat)
                ax10.bar(x + multiplier*barWidth, list(totIns.values()), width=barWidth, label=mat)

            multiplier += 1
        
        print("Plotting images")
        # 1. Figure saving raceTime Mean Max graph
        ax1.set_title(f"{nRow}: Race time Max Mean comparison ")
        ax1.set_xticks(x+barWidth*2, raceTimeMaxMean.keys(), rotation="vertical")
        ax1.grid(axis='y')
        ax1.legend()
        fig1.savefig(os.path.join(postPath, f"{nRow}-raceTimeMeanMax_comparison.png"), bbox_inches='tight')

        # 2. Figure saving raceTime Max Max graph
        ax2.set_title(f"{nRow}: Race time Max Max comparison ")
        ax2.set_xticks(x+barWidth*2, raceTimeMaxMax.keys(), rotation="vertical")
        ax2.grid(axis='y')
        ax2.legend()
        fig2.savefig(os.path.join(postPath, f"{nRow}-raceTimeMaxMax_comparison.png"), bbox_inches='tight')

        # 3. Figure saving warmUpTime Mean Max graph
        ax3.set_title(f"{nRow}: Warm up time Max Mean comparison ")
        ax3.set_xticks(x+barWidth*2, warmUpTimeMaxMean.keys(), rotation="vertical")
        ax3.grid(axis='y')
        ax3.legend()
        fig3.savefig(os.path.join(postPath, f"{nRow}-warmUpTimeMeanMax_comparison.png"), bbox_inches='tight')

        # 4. Figure saving warmUpTime Max Max graph
        ax4.set_title(f"{nRow}: Warm up time Max Max comparison ")
        ax4.set_xticks(x+barWidth*2, warmUpTimeMaxMax.keys(), rotation="vertical")
        ax4.grid(axis='y')
        ax4.legend()
        fig4.savefig(os.path.join(postPath, f"{nRow}-warmUpTimeMaxMax_comparison.png"), bbox_inches='tight')

        if instr == 1:
            ax5.set_title(f"{nRow}: stall on memory writing comparison")
            ax5.set_xticks(x+barWidth*2, stlMemWrt.keys(), rotation="vertical")
            ax5.grid(axis='y')
            ax5.legend()
            fig5.savefig(os.path.join(postPath, f"{nRow}-stlMemWrt_comparison.png"), bbox_inches='tight')

            ax6.set_title(f"{nRow}: stall on any resources comparison")
            ax6.set_xticks(x+barWidth*2, stlRes.keys(), rotation="vertical")
            ax6.grid(axis='y')
            ax6.legend()
            fig6.savefig(os.path.join(postPath, f"{nRow}-stlRes_comparison.png"), bbox_inches='tight')

            ax7.set_title(f"{nRow}: Total cycles comparison")
            ax7.set_xticks(x+barWidth*2, totCyc.keys(), rotation="vertical")
            ax7.grid(axis='y')
            ax7.legend()
            fig7.savefig(os.path.join(postPath, f"{nRow}-totCyc_comparison.png"), bbox_inches='tight')

            ax8.set_title(f"{nRow}: double precision operations comparison ")
            ax8.set_xticks(x+barWidth*2, dpOps.keys(), rotation="vertical")
            ax8.grid(axis='y')
            ax8.legend()
            fig8.savefig(os.path.join(postPath, f"{nRow}-dpOps_comparison.png"), bbox_inches='tight')

            ax8.set_title(f"{nRow}: double precision vectorial operations comparison ")
            ax8.set_xticks(x+barWidth*2, dpVecOps.keys(), rotation="vertical")
            ax8.grid(axis='y')
            ax8.legend()
            fig8.savefig(os.path.join(postPath, f"{nRow}-dpVecOps_comparison.png"), bbox_inches='tight')

        if instr == 2:
            ax5.set_title(f"{nRow}: L1 data cache misses comparison ")
            ax5.set_xticks(x+barWidth*2, l1dcm.keys(), rotation="vertical")
            ax5.grid(axis='y')
            ax5.legend()
            fig5.savefig(os.path.join(postPath, f"{nRow}-L1DCM_comparison.png"), bbox_inches='tight')

            ax6.set_title(f"{nRow}: L1 instraction cache misses comparison ")
            ax6.set_xticks(x+barWidth*2, l1icm.keys(), rotation="vertical")
            ax6.grid(axis='y')
            ax6.legend()
            fig6.savefig(os.path.join(postPath, f"{nRow}-L1ICM_comparison.png"), bbox_inches='tight')

            ax7.set_title(f"{nRow}: L2 data cache misses comparison ")
            ax7.set_xticks(x+barWidth*2, l2dcm.keys(), rotation="vertical")
            ax7.grid(axis='y')
            ax7.legend()
            fig7.savefig(os.path.join(postPath, f"{nRow}-L2DCM_comparison.png"), bbox_inches='tight')

            ax8.set_title(f"{nRow}: L2 instruction cache misses comparison ")
            ax8.set_xticks(x+barWidth*2, l2icm.keys(), rotation="vertical")
            ax8.grid(axis='y')
            ax8.legend()
            fig8.savefig(os.path.join(postPath, f"{nRow}-L2ICM_comparison.png"), bbox_inches='tight')

            ax9.set_title(f"{nRow}: L3 total cache misses comparison ")
            ax9.set_xticks(x+barWidth*2, l3tcm.keys(), rotation="vertical")
            ax9.grid(axis='y')
            ax9.legend()
            fig9.savefig(os.path.join(postPath, f"{nRow}-L3TCM_comparison.png"), bbox_inches='tight')

        if instr == 3:
            ax5.set_title(f"{nRow}: loaded instruction comparison ")
            ax5.set_xticks(x+barWidth*2, ldIns.keys(), rotation="vertical")
            ax5.grid(axis='y')
            ax5.legend()
            fig5.savefig(os.path.join(postPath, f"{nRow}-ldIns_comparison.png"), bbox_inches='tight')

            ax6.set_title(f"{nRow}: stored instraction comparison ")
            ax6.set_xticks(x+barWidth*2, stIns.keys(), rotation="vertical")
            ax6.grid(axis='y')
            ax6.legend()
            fig6.savefig(os.path.join(postPath, f"{nRow}-stIns_comparison.png"), bbox_inches='tight')

            ax7.set_title(f"{nRow}: branch instructions comparison ")
            ax7.set_xticks(x+barWidth*2, brIns.keys(), rotation="vertical")
            ax7.grid(axis='y')
            ax7.legend()
            fig7.savefig(os.path.join(postPath, f"{nRow}-brIns_comparison.png"), bbox_inches='tight')

            ax8.set_title(f"{nRow}: branch conditional instructions comparison ")
            ax8.set_xticks(x+barWidth*2, brConIns.keys(), rotation="vertical")
            ax8.grid(axis='y')
            ax8.legend()
            fig8.savefig(os.path.join(postPath, f"{nRow}-brConIns_comparison.png"), bbox_inches='tight')

            ax9.set_title(f"{nRow}: branch miss predition comparison ")
            ax9.set_xticks(x+barWidth*2, brMiss.keys(), rotation="vertical")
            ax9.grid(axis='y')
            ax9.legend()
            fig9.savefig(os.path.join(postPath, f"{nRow}-brMiss_comparison.png"), bbox_inches='tight')

            ax10.set_title(f"{nRow}: total instruction comparison ")
            ax10.set_xticks(x+barWidth*2, totIns.keys(), rotation="vertical")
            ax10.grid(axis='y')
            ax10.legend()
            fig10.savefig(os.path.join(postPath, f"{nRow}-totIns_comparison.png"), bbox_inches='tight')

        plt.close()




                
                


